---====================== Start Software License ======================---
--==                                                                  ==--
--== This license governs the use of this software, and your use of   ==--
--== this software constitutes acceptance of this license. Agreement  ==--
--== with all points is required to use this software.                ==--
--==                                                                  ==--
--== 1. You may use this software freely for personal use.            ==--
--==                                                                  ==--
--== 2. You may use this software freely to determine feasibility for ==--
--==    commercial use.                                               ==--
--==                                                                  ==--
--== 3. You may use this software for commercial use if the author    ==--
--==    has given you written consent.                                ==--
--==                                                                  ==--
--== 4. You may modify this software provided you do not remove the   ==--
--==    license and copyright notice.                                 ==--
--==                                                                  ==--
--== 5. You may distribute this software and derivative works to      ==--
--==    personal friends and work colleagues only.                    ==--
--==                                                                  ==--
--== 6. You agree that this software comes as-is and with no        ==--
--==    warranty whatsoever, either expressed or implied, including,  ==--
--==    but not limited to, warranties of merchantability or fitness  ==--
--==    for a particular purpose.                                     ==--
--==                                                                  ==--
--== 7. You agree that the author will not be liable for any damages  ==--
--==    relating from the use of this software, including direct,     ==--
--==    indirect, consequential or incidental. This software is used  ==--
--==    entirely at your own risk and should it prove defective, you  ==--
--==    will assume full responsibility for all costs associated with ==--
--==    servicing, repair or correction.                              ==--
--==                                                                  ==--
--== Your rights under this license are terminated immediately if you ==--
--== breach it in any way.                                            ==--
--==                                                                  ==--
---======================= End Software License =======================---


---====================== Start Copyright Notice ======================---
--==                                                                  ==--
--== Filename ..... raptor16.vhd                                      ==--
--== Download ..... http://www.spacewire.co.uk                        ==--
--== Author ....... Steve Haywood (steve.haywood@ukonline.co.uk)      ==--
--== Copyright .... Copyright (c) 2004 Steve Haywood                  ==--
--== Project ...... Raptor-16                                         ==--
--== Version ...... 1.00                                              ==--
--== Conception ... 10 June 2004                                      ==--
--== Modified ..... N/A                                               ==--
--==                                                                  ==--
---======================= End Copyright Notice =======================---


---========================= Start Description ========================---
--==                                                                  ==--
--== A full 16-bit CISC Microprocessor - see external documentation.  ==--
--==                                                                  ==--
---========================== End Description =========================---


LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;


ENTITY raptor16 IS
  PORT(
       --== General Interface (Sync Rst, 50MHz Clock) ==--

       clk         : IN  STD_LOGIC;
       rst         : IN  STD_LOGIC;

       --== User Port 0 Interface ==--

       port0_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port0_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port0_read  : OUT STD_LOGIC;
       port0_write : OUT STD_LOGIC;

       --== User Port 1 Interface ==--

       port1_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port1_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port1_read  : OUT STD_LOGIC;
       port1_write : OUT STD_LOGIC;

       --== User Port 2 Interface ==--

       port2_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port2_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port2_read  : OUT STD_LOGIC;
       port2_write : OUT STD_LOGIC;

       --== User Port 3 Interface ==--

       port3_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port3_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port3_read  : OUT STD_LOGIC;
       port3_write : OUT STD_LOGIC;

       --== User Port 4 Interface ==--

       port4_din   : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       port4_dout  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       port4_read  : OUT STD_LOGIC;
       port4_write : OUT STD_LOGIC;

       --== Interrupt Priority 1 (Lowest) Interface ==--

       int_req1    : IN  STD_LOGIC;
       int_ack1    : OUT STD_LOGIC;

       --== Interrupt Priority 2 Interface ==--

       int_req2    : IN  STD_LOGIC;
       int_ack2    : OUT STD_LOGIC;

       --== Interrupt Priority 3 Interface ==--

       int_req3    : IN  STD_LOGIC;
       int_ack3    : OUT STD_LOGIC;

       --== Interrupt Priority 4 Interface ==--

       int_req4    : IN  STD_LOGIC;
       int_ack4    : OUT STD_LOGIC;

       --== Interrupt Priority 5 Interface ==--

       int_req5    : IN  STD_LOGIC;
       int_ack5    : OUT STD_LOGIC;

       --== Interrupt Priority 6 Interface ==--

       int_req6    : IN  STD_LOGIC;
       int_ack6    : OUT STD_LOGIC;

       --== Interrupt Priority 7 (Highest) Interface ==--

       int_req7    : IN  STD_LOGIC;
       int_ack7    : OUT STD_LOGIC;

       --== Static RAM Interface ==--

       ram_en      : OUT STD_LOGIC;
       ram_we      : OUT STD_LOGIC;
       ram_addr    : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
       ram_dout    : IN  STD_LOGIC_VECTOR(15 DOWNTO 0);
       ram_din     : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)
      );
END raptor16;


ARCHITECTURE rtl OF raptor16 IS

---=====================================---
--== Instruction Constant Declarations ==--
---=====================================---

CONSTANT inst_move : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0000";
CONSTANT inst_and  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0001";
CONSTANT inst_or   : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0010";
CONSTANT inst_xor  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0011";
CONSTANT inst_shl  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0100";
CONSTANT inst_shr  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0101";
CONSTANT inst_add  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0110";
CONSTANT inst_sub  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0111";
CONSTANT inst_cmp  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1000";
CONSTANT inst_mulu : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1001"; -- TBD
CONSTANT inst_muls : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1010"; -- TBD
CONSTANT inst_divu : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1011"; -- TBD
CONSTANT inst_divs : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1100"; -- TBD
CONSTANT inst_bra  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1101";
CONSTANT inst_bsr  : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1110";
CONSTANT inst_load : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1111";

---=========================================---
--== Addressing Mode Constant Declarations ==--
---=========================================---

CONSTANT mode_dir_dreg  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "000";
CONSTANT mode_dir_areg  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "001";
CONSTANT mode_indir     : STD_LOGIC_VECTOR(2 DOWNTO 0) := "010";
CONSTANT mode_indir_pd  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "011";
CONSTANT mode_indir_pi  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "100";
CONSTANT mode_indir_dsp : STD_LOGIC_VECTOR(2 DOWNTO 0) := "101";
CONSTANT mode_imm_quick : STD_LOGIC_VECTOR(2 DOWNTO 0) := "110";
CONSTANT mode_extend    : STD_LOGIC_VECTOR(2 DOWNTO 0) := "111";

CONSTANT mode_ext_imm   : STD_LOGIC_VECTOR(2 DOWNTO 0) := "110";
CONSTANT mode_ext_addr  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "111";

---==========================================---
--== Branch Condition Constant Declarations ==--
---==========================================---

CONSTANT branch_alw : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0000";
CONSTANT branch_nev : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0001";
CONSTANT branch_zrs : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0010";
CONSTANT branch_zrc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0011";
CONSTANT branch_ngs : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0100";
CONSTANT branch_ngc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0101";
CONSTANT branch_uos : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0110";
CONSTANT branch_uoc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "0111";
CONSTANT branch_sos : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1000";
CONSTANT branch_soc : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1001";
CONSTANT branch_ugt : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1010";
CONSTANT branch_ule : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1011";
CONSTANT branch_sgt : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1100";
CONSTANT branch_sge : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1101";
CONSTANT branch_slt : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1110";
CONSTANT branch_sle : STD_LOGIC_VECTOR(3 DOWNTO 0) := "1111";

CONSTANT branch_equ : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_zrs;
CONSTANT branch_neq : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_zrc;
CONSTANT branch_ult : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_uos;
CONSTANT branch_uge : STD_LOGIC_VECTOR(3 DOWNTO 0) := branch_uoc;

---=====================================---
--== Mux Control Constant Declarations ==--
---=====================================---

CONSTANT srdsp_one    : STD_LOGIC := '0'; -- sel_reg_disp Constant
CONSTANT srdsp_ram    : STD_LOGIC := '1'; -- sel_reg_disp Constant

CONSTANT sdsub_add    : STD_LOGIC := '0'; -- sel_disp_sub Constant
CONSTANT sdsub_sub    : STD_LOGIC := '1'; -- sel_disp_sub Constant

CONSTANT shreg_ram    : STD_LOGIC := '0'; -- sel_hold_reg Constant
CONSTANT shreg_dsp    : STD_LOGIC := '1'; -- sel_hold_reg Constant

CONSTANT sraddr_pc    : STD_LOGIC_VECTOR(1 DOWNTO 0) := "11"; -- sel_ram_addr Constant
CONSTANT sraddr_hold  : STD_LOGIC_VECTOR(1 DOWNTO 0) := "01"; -- sel_ram_addr Constant
CONSTANT sraddr_sp    : STD_LOGIC_VECTOR(1 DOWNTO 0) := "10"; -- sel_ram_addr Constant
CONSTANT sraddr_reg   : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00"; -- sel_ram_addr Constant

CONSTANT sfeed_reg    : STD_LOGIC := '0'; -- sel_feedback Constant
CONSTANT sfeed_res    : STD_LOGIC := '1'; -- sel_feedback Constant

CONSTANT srdin_res    : STD_LOGIC := '0'; -- sel_ram_din Constant
CONSTANT srdin_pc     : STD_LOGIC := '1'; -- sel_ram_din Constant

CONSTANT sopsrc_ram   : STD_LOGIC_VECTOR(2 DOWNTO 0) := "000"; -- sel_op_source Constant
CONSTANT sopsrc_port  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "001"; -- sel_op_source Constant
CONSTANT sopsrc_dreg  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "010"; -- sel_op_source Constant
CONSTANT sopsrc_areg  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "011"; -- sel_op_source Constant
CONSTANT sopsrc_immq  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "100"; -- sel_op_source Constant
CONSTANT sopsrc_braq  : STD_LOGIC_VECTOR(2 DOWNTO 0) := "101"; -- sel_op_source Constant
CONSTANT sopsrc_immqp : STD_LOGIC_VECTOR(2 DOWNTO 0) := "110"; -- sel_op_source Constant
CONSTANT sopsrc_immqn : STD_LOGIC_VECTOR(2 DOWNTO 0) := "111"; -- sel_op_source Constant

---=====================---
--== Type Declarations ==--
---=====================---

TYPE REG_BLOCK IS ARRAY(0 TO 7) OF STD_LOGIC_VECTOR(15 DOWNTO 0);
TYPE PORT_BLOCK IS ARRAY(0 TO 4) OF STD_LOGIC_VECTOR(15 DOWNTO 0);

TYPE STATES IS
  (st_decode,
   st_dec_reg,
   st_fetch_operand,
   st_fetch_data,
   st_save_data,
   st_save_operand,
   st_fetch_execute,
   st_dec_sp,
   st_save_writeback
  );

TYPE INT_STATES IS
  (int_none,
   int_pop_cr,
   int_push_cr,
   int_push_pc,
   int_push,
   int_pop
  );

---===================================---
--== Mux Control Signal Declarations ==--
---===================================---

SIGNAL sel_reg_disp  : STD_LOGIC;
SIGNAL sel_disp_sub  : STD_LOGIC;
SIGNAL sel_hold_reg  : STD_LOGIC;
SIGNAL sel_ram_addr  : STD_LOGIC_VECTOR(1 DOWNTO 0);
SIGNAL sel_feedback  : STD_LOGIC;
SIGNAL sel_ram_din   : STD_LOGIC;
SIGNAL sel_op_source : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL sel_register  : STD_LOGIC_VECTOR(3 DOWNTO 0);
SIGNAL sel_port      : STD_LOGIC_VECTOR(2 DOWNTO 0);

---======================================---
--== Register Write Signal Declarations ==--
---======================================---

SIGNAL register_we : STD_LOGIC;
SIGNAL port_we     : STD_LOGIC;
SIGNAL src_op_we   : STD_LOGIC;
SIGNAL dst_op_we   : STD_LOGIC;
SIGNAL hold_we     : STD_LOGIC;
SIGNAL result_we   : STD_LOGIC;

---======================================---
--== Condition Code Signal Declarations ==--
---======================================---

SIGNAL zr_i : STD_LOGIC;
SIGNAL ng_i : STD_LOGIC;
SIGNAL so_i : STD_LOGIC;
SIGNAL uo_i : STD_LOGIC;
SIGNAL zr_m : STD_LOGIC;
SIGNAL ng_m : STD_LOGIC;
SIGNAL so_m : STD_LOGIC;
SIGNAL uo_m : STD_LOGIC;
SIGNAL zr   : STD_LOGIC;
SIGNAL ng   : STD_LOGIC;
SIGNAL uo   : STD_LOGIC;
SIGNAL so   : STD_LOGIC;

---======================================---
--== Processor Core Signal Declarations ==--
---======================================---

SIGNAL port_registers  : PORT_BLOCK;
SIGNAL port_read       : STD_LOGIC_VECTOR(4 DOWNTO 0);
SIGNAL port_write      : STD_LOGIC_VECTOR(4 DOWNTO 0);
SIGNAL dregs           : REG_BLOCK;
SIGNAL aregs           : REG_BLOCK;
SIGNAL cur_dreg        : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL cur_areg        : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
SIGNAL cur_port_reg    : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL displaced_op    : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
SIGNAL displaced       : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL src_op          : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
SIGNAL dst_op          : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
SIGNAL result_i        : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL result          : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL op_source       : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL feedback        : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL hold_reg        : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL hold_reg_d      : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL result_shl      : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL result_shr      : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL result_we_delay : STD_LOGIC;
SIGNAL con_reg         : STD_LOGIC_VECTOR(15 DOWNTO 0);

---=============================---
--== Alias Signal Declarations ==--
---=============================---

SIGNAL pre_inst  : STD_LOGIC_VECTOR(3 DOWNTO 0);
SIGNAL inst      : STD_LOGIC_VECTOR(3 DOWNTO 0);
SIGNAL condition : STD_LOGIC_VECTOR(3 DOWNTO 0);
SIGNAL src_mode  : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL src_data  : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL dst_mode  : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL dst_data  : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL pc        : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL sp        : STD_LOGIC_VECTOR(15 DOWNTO 0);

---=========================================---
--== Processor Control Signal Declarations ==--
---=========================================---

SIGNAL state            : STATES;
SIGNAL operand          : STD_LOGIC;
SIGNAL instruction      : STD_LOGIC_VECTOR(15 DOWNTO 0);
SIGNAL mode             : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL data             : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL execute          : STD_LOGIC;
SIGNAL block_conditions : STD_LOGIC;
SIGNAL bsr              : STD_LOGIC;
SIGNAL fetch            : STD_LOGIC;
SIGNAL inst_eq_move     : STD_LOGIC;
SIGNAL inst_eq_cmp      : STD_LOGIC;
SIGNAL inst_eq_bra_bsr  : STD_LOGIC;
SIGNAL bad_inst         : STD_LOGIC;
SIGNAL quick_data_ext   : STD_LOGIC_VECTOR(4 DOWNTO 0);
SIGNAL reg_mask         : STD_LOGIC_VECTOR(6 DOWNTO 0);
SIGNAL reg_num          : STD_LOGIC_VECTOR(2 DOWNTO 0);

---=================================---
--== Interrupt Signal Declarations ==--
---=================================---

SIGNAL int_stage       : INT_STATES;
SIGNAL int_level       : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL int_we          : STD_LOGIC;
SIGNAL save_int_level  : STD_LOGIC;
SIGNAL int_ack_i       : STD_LOGIC_VECTOR(7 DOWNTO 1);
SIGNAL early_int_check : STD_LOGIC;

BEGIN

  ---=====================---
  --== Alias Assignments ==--
  ---=====================---

  pre_inst  <= ram_dout(15 DOWNTO 12);
  condition <= ram_dout(11 DOWNTO 8);
  inst      <= instruction(15 DOWNTO 12);
  dst_mode  <= instruction(11 DOWNTO 9);
  dst_data  <= instruction(8 DOWNTO 6);
  src_mode  <= instruction(5 DOWNTO 3);
  src_data  <= instruction(2 DOWNTO 0);
  pc        <= aregs(7);
  sp        <= aregs(6);
  so        <= con_reg(11);
  uo        <= con_reg(10);
  ng        <= con_reg(09);
  zr        <= con_reg(08);

  ---=====================================---
  --== Operand Addressing Mode Selection ==--
  ---=====================================---

  PROCESS(operand, src_mode, src_data, dst_mode, dst_data)
  BEGIN
    CASE operand IS
      WHEN '0'    => mode <= src_mode;
                     data <= src_data;
      WHEN '1'    => mode <= dst_mode;
                     data <= dst_data;
      WHEN OTHERS => NULL;
    END CASE;
  END PROCESS;

  ---===============================---
  --== Operand Write Control Logic ==--
  ---===============================---

  src_op_we <= '1' WHEN (state = st_decode) AND (operand = '1') ELSE '0';
  dst_op_we <= '1' WHEN (state = st_save_operand) ELSE '0';

  ---===========================---
  --== Processor Control Logic ==--
  ---===========================---

  PROCESS(clk)

  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN

        sel_feedback     <= '0';
        register_we      <= '0';
        sel_register     <= (OTHERS => '0');
        sel_reg_disp     <= '0';
        sel_disp_sub     <= '0';
        sel_ram_addr     <= "00";
        ram_en           <= '0';
        ram_we           <= '0';
        sel_op_source    <= (OTHERS => '0');
        hold_we          <= '0';
        sel_hold_reg     <= '0';
        port_we          <= '0';
        sel_port         <= (OTHERS => '0');
        state            <= st_save_operand;
        instruction      <= (OTHERS => '0');
        operand          <= '0';
        result_we        <= '0';
        sel_ram_din      <= '0';
        block_conditions <= '0';
        bsr              <= '0';
        fetch            <= '0';
        inst_eq_move     <= '0';
        inst_eq_cmp      <= '0';
        bad_inst         <= '0';
        inst_eq_bra_bsr  <= '0';
        quick_data_ext   <= (OTHERS => '0');
        save_int_level   <= '0';
        int_stage        <= int_none;
        int_we           <= '0';
        reg_mask         <= (OTHERS => '0');
        reg_num          <= (OTHERS => '0');
        early_int_check  <= '0';

      ELSE

        --
        -- Default Signal Assigns.
        --
        sel_reg_disp  <= srdsp_one;
        sel_disp_sub  <= sdsub_add;
        sel_hold_reg  <= shreg_ram;
        sel_ram_addr  <= sraddr_reg;
        sel_feedback  <= sfeed_reg;
        sel_ram_din   <= srdin_res;
        sel_op_source <= sopsrc_ram;

        register_we   <= '0';
        ram_en        <= '0';
        ram_we        <= '0';
        hold_we       <= '0';
        port_we       <= '0';
        result_we     <= '0';
        int_we        <= '0';

        --
        -- Port Selection.
        --
        sel_port <= data(2 DOWNTO 0);

        --
        -- Register Selection.
        --
        CASE mode IS
          WHEN mode_dir_dreg =>
            sel_register <= '0' & data;
          WHEN OTHERS =>
            sel_register <= '1' & data;
        END CASE;


        --
        -- Processor State Machine.
        --
        CASE state IS

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An    Yes  Yes
          -- (An)     Yes  Yes
          -- -(An)    Yes  Yes
          -- (An)+    Yes  Yes
          -- d(An)    Yes  Yes
          -- ImmQ     Yes   -
          -- Pn       Yes  Yes
          -- Imm      Yes   -
          -- Addr     Yes  Yes
          --
          WHEN st_decode =>

            --
            -- Disable Condition Code update for any instruction that modifies
            -- the Program Counter. Also disable it for invalid instructions.
            --
            IF ((dst_mode = mode_dir_areg) AND (dst_data = "111") AND
                (inst_eq_cmp = '0')) OR (bad_inst = '1') THEN
              block_conditions <= '1';
            ELSE
              block_conditions <= '0';
            END IF;

            sel_reg_disp <= srdsp_one; -- *MUST* for -(An), (An)+ [Src Op], d(An), Imm & Addr

            IF (mode = mode_indir_pd) THEN
              sel_disp_sub <= sdsub_sub; -- *MUST* for -(An)
            ELSE
              sel_disp_sub <= sdsub_add; -- *MUST* for (An)+ [Src Op], d(An), Imm & Addr
            END IF;

            sel_feedback <= sfeed_reg; -- *MUST* for -(An), (An)+, d(An), Imm & Addr

            IF (mode = mode_indir) OR (mode = mode_indir_pi) THEN
              sel_ram_addr <= sraddr_reg; -- *MUST* for (An) & (An)+
            ELSE
              sel_ram_addr  <= sraddr_pc; -- *MUST* for d(An), Imm & Addr
            END IF;

            IF NOT(inst = inst_load) AND ((mode = mode_indir_dsp) OR
               ((mode = mode_extend) AND ((data = mode_ext_imm) OR (data = mode_ext_addr)))) THEN
              sel_register <= "1111"; -- *MUST BE* "1111" for d(An), Imm & Addr
            END IF;

            IF (bad_inst = '1') THEN
              state <= st_save_operand;
            ELSE

              CASE mode IS

                WHEN mode_dir_dreg | mode_dir_areg => -- Dn/An

                  sel_op_source <= '0' & '1' & mode(0);

                  IF (operand = '0') THEN
                    operand <= '1';
                    IF (execute = '0') THEN
                      --
                      -- Skip Src & Dst Op Fetch (1 cycles saved for Dn/An)
                      --
                      state <= st_save_operand;
                    ELSE
                      state <= st_decode;
                    END IF;
                  ELSE
                    state <= st_save_operand;
                  END IF;

                WHEN mode_indir => -- (An)

                  IF ((operand = '1') AND (inst_eq_move = '1')) OR
                     ((operand = '0') AND (execute = '0')) THEN
                    --
                    -- Skip Dst Op Fetch (1 cycle saved for (An)/(An)+)
                    -- Skip Src & Dst Op Fetch (2 cycles saved for (An)/(An)+)
                    --
                    state <= st_save_operand;
                  ELSE
                    state <= st_fetch_operand;
                  END IF;
                  ram_en <= '1'; -- *MUST BE* '1' for (An), (An)+, d(An), Imm & Addr

                WHEN mode_indir_pi => -- (An)+

                  IF ((operand = '1') AND (inst_eq_move = '1')) OR
                     ((operand = '0') AND (execute = '0')) THEN
                    --
                    -- Skip Move Dst Op Fetch (1 cycle saved for (An)/(An)+)
                    -- Skip Bra/Bsr/Pop Src & Dst Op Fetch (2 cycles saved for (An)/(An)+)
                    --
                    state <= st_save_operand;
                  ELSE
                    state <= st_fetch_operand;
                  END IF;
                  ram_en <= '1'; -- *MUST BE* '1' for (An), (An)+, d(An), Imm & Addr
                  --
                  -- No Post-Inc when Operand = Src Op or Bra/Bsr Not Taken or No Reg Pop.
                  --
                  IF (operand = '0') AND (execute = '1') THEN
                    register_we <= '1'; -- *MUST BE* '1' for -(An), (An)+ [Src Op], d(An), Imm & Addr
                  END IF;

                WHEN mode_indir_pd => -- -(An)

                  IF ((operand = '1') AND (inst_eq_move = '1')) OR
                     ((operand = '0') AND (execute = '0')) THEN
                    --
                    -- Skip Move Dst Op Fetch (2 cycles saved for -(An))
                    -- Skip Bra/Bsr/Push Src & Dst Op Fetch (3 cycles saved for -(An))
                    --
                    state <= st_save_operand;
                  ELSE
                    state <= st_dec_reg;
                  END IF;

                  --
                  -- No Pre-Dec when Bra/Bsr Not Taken or No Reg Push.
                  --
                  IF (execute = '1') THEN
                    register_we <= '1'; -- *MUST BE* '1' for -(An), (An)+ [Src Op], d(An), Imm & Addr
                  END IF;

                WHEN mode_imm_quick => -- Imm (Quick)
                  --
                  -- All Quick Instructions Come Here (BraQ, BsrQ, LoadQ, InstQ & InstQ (Ext))
                  --
                  IF (operand = '0') AND (execute = '0') THEN
                    --
                    -- Skip Src & Dst Op Fetch (1 cycles saved for ImmQ)
                    --
                    state <= st_save_operand;
                  ELSE
                    state <= st_decode;
                  END IF;
                  sel_op_source <= sopsrc_immq;
                  operand <= '1';

                WHEN mode_indir_dsp =>

                  state <= st_fetch_data;
                  ram_en <= '1'; -- *MUST BE* '1' for (An), (An)+, d(An), Imm & Addr
                  register_we <= '1'; -- *MUST BE* '1' for -(An), (An)+ [Src Op], d(An), Imm & Addr

                WHEN OTHERS =>

                  CASE data IS

                    WHEN mode_ext_imm | mode_ext_addr =>

                      state <= st_fetch_data;
                      ram_en <= '1'; -- *MUST BE* '1' for (An), (An)+, d(An), Imm & Addr
                      register_we <= '1'; -- *MUST BE* '1' for -(An), (An)+ [Src Op], d(An), Imm & Addr

                    WHEN OTHERS =>

                      sel_op_source <= sopsrc_port;
                      IF (operand = '0') THEN
                        operand <= '1';
                        IF (execute = '0') THEN
                          --
                          -- Skip Src & Dst Op Fetch (1 cycle saved for Pn)
                          --
                          state <= st_save_operand;
                        ELSE
                          state <= st_decode;
                        END IF;
                      ELSE
                        state <= st_save_operand;
                      END IF;

                  END CASE;

              END CASE;

            END IF;

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An     -    -
          -- (An)      -    -
          -- -(An)    Yes  Yes
          -- (An)+     -    -
          -- d(An)     -    -
          -- ImmQ      -    -
          -- Pn        -    -
          -- Imm       -    -
          -- Addr      -    -
          --
          WHEN st_dec_reg =>

            sel_ram_addr <= sraddr_reg;
            ram_en <= '1';
            state <= st_fetch_operand;

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An     -    -
          -- (An)      -    -
          -- -(An)    Yes  Yes
          -- (An)+    Yes  Yes
          -- d(An)    Yes  Yes
          -- ImmQ      -    -
          -- Pn        -    -
          -- Imm       -    -
          -- Addr      -    -
          --
          WHEN st_fetch_operand =>

            sel_op_source <= sopsrc_ram;
            IF (operand = '0') THEN
              operand <= '1';
              state <= st_decode;
            ELSE
              state <= st_save_operand;
            END IF;

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An     -    -
          -- (An)      -    -
          -- -(An)     -    -
          -- (An)+     -    -
          -- d(An)    Yes  Yes
          -- ImmQ      -    -
          -- Pn        -    -
          -- Imm      Yes   -
          -- Addr     Yes  Yes
          --
          WHEN st_fetch_data =>

            sel_op_source <= sopsrc_ram; --*MUST* for Imm
            sel_reg_disp <= srdsp_ram; -- *MUST* for d(An)
            sel_disp_sub <= sdsub_add; -- *MUST* for d(An)

            IF (mode = mode_indir_dsp) THEN
              sel_hold_reg <= shreg_dsp; -- *MUST* for d(An)
            ELSE
              sel_hold_reg <= shreg_ram; -- *MUST* for Addr
            END IF;

            IF ((operand = '1') AND (inst_eq_move = '1')) OR
               ((operand = '0') AND (execute = '0')) THEN
              --
              -- Skip Dst Op Fetch (2 cycles saved for d(An)/Addr)
              -- Skip Src & Dst Op Fetch (3 cycles saved for d(An)/Addr)
              --
              state <= st_save_operand;
            ELSIF (mode = mode_extend) AND (data = mode_ext_imm) THEN
              operand <= '1';
              state <= st_decode;
            ELSE
              state <= st_save_data;
            END IF;

            IF NOT((mode = mode_extend) AND (data = mode_ext_imm)) THEN
              hold_we <= '1'; -- *MUST BE* for d(An) & Addr
            END IF;

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An     -    -
          -- (An)      -    -
          -- -(An)     -    -
          -- (An)+     -    -
          -- d(An)     -   Yes
          -- ImmQ      -    -
          -- Pn        -    -
          -- Imm       -    -
          -- Addr      -   Yes
          --
          WHEN st_save_data =>

            sel_ram_addr <= sraddr_hold;
            ram_en <= '1';
            state <= st_fetch_operand;

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An     -   Yes
          -- (An)      -   Yes
          -- -(An)     -   Yes
          -- (An)+     -   Yes
          -- d(An)     -    -
          -- ImmQ      -    -
          -- Pn        -   Yes
          -- Imm       -    -
          -- Addr      -    -
          --
          WHEN st_save_operand =>

            sel_feedback <= sfeed_reg;
            sel_ram_addr <= sraddr_pc;
            sel_reg_disp <= srdsp_one;

            IF (bad_inst = '0') AND (execute = '1') AND (data = "111") AND
               ((mode = mode_dir_areg) OR (mode = mode_indir_pi)) THEN  -- Execute Last Instruction & Delay Fetching of Next

              sel_register <= "1110";
              sel_disp_sub <= sdsub_sub;
              fetch <= '0';
              IF (bsr = '1') THEN
                register_we <= '1';
                state <= st_dec_sp;
              ELSE
                result_we <= '1';
                state <= st_fetch_execute;
              END IF;

              --
              -- Interrupt Detection.
              --
              IF (int_stage = int_none) THEN
                IF (early_int_check = '1') THEN
                  IF (int_level /= "000") THEN
                    int_stage <= int_push_cr;
                  END IF;
                ELSIF (int_level > con_reg(14 DOWNTO 12)) THEN
                  int_stage <= int_push_cr;
                END IF;
              END IF;

            ELSE

              ram_en <= '1';
              sel_register <= "1111";
              sel_disp_sub <= sdsub_add;
              register_we <= '1';
              IF (execute = '1') AND (bad_inst = '0') THEN
                result_we <= '1'; -- Execute Last Instruction & Fetch Next 
              ELSE
                result_we <= '0'; -- Skip Executing Last Instruction & Fetch Next
              END IF;
              fetch <= '1';
              state <= st_fetch_execute;

              --
              -- Interrupt Detection.
              --
              IF (int_stage = int_none) THEN
                IF (early_int_check = '1') THEN
                  IF (int_level /= "000") THEN
                    int_stage <= int_push_cr;
                    fetch <= '0';
                    ram_en <= '0';
                    register_we <= '0';
                  END IF;
                ELSIF (int_level > con_reg(14 DOWNTO 12)) THEN
                  fetch <= '0';
                  ram_en <= '0';
                  register_we <= '0';
                  int_stage <= int_push_cr;
                END IF;
              ELSE
                fetch <= '0';
                ram_en <= '0';
                register_we <= '0';
              END IF;

            END IF;

          WHEN st_dec_sp =>

            sel_ram_addr <= sraddr_sp;
            sel_ram_din <= srdin_pc;
            ram_en <= '1';
            ram_we <= '1';
            result_we <= '1';
            state <= st_fetch_execute;

          --
          -- Called with the following Addressing Modes:-
          --
          -- Mode     Src  Dst
          -- =================
          -- Dn/An     -   Yes
          -- (An)      -   Yes
          -- -(An)     -   Yes
          -- (An)+     -   Yes
          -- d(An)     -   Yes
          -- ImmQ      -    -
          -- Pn        -   Yes
          -- Imm       -    -
          -- Addr      -   Yes
          --
          WHEN st_fetch_execute =>

            sel_ram_din <= srdin_res; -- *MUST* for (An), -(An), (An)+, d(An) & Addr
            sel_reg_disp <= srdsp_one; -- *MUST* for (An)+
            sel_disp_sub <= sdsub_add; -- *MUST* for (An)+

            IF (mode = mode_dir_dreg) OR (mode = mode_dir_areg) THEN
              sel_feedback <= sfeed_res; -- *MUST* for Dn/An
            ELSE
              sel_feedback <= sfeed_reg; -- *MUST* for (An)+
            END IF;

            IF (mode = mode_indir) OR (mode = mode_indir_pd) OR (mode = mode_indir_pi) THEN
              sel_ram_addr <= sraddr_reg; -- *MUST* for (An), -(An) & (An)+
            ELSE
              sel_ram_addr <= sraddr_hold; -- *MUST* for d(An) & Addr
            END IF;

            IF (inst_eq_cmp = '0') AND (execute = '1') AND (bad_inst = '0') THEN

              IF (mode = mode_dir_dreg) OR (mode = mode_dir_areg) OR (mode = mode_indir_pi) THEN
                register_we <= '1'; -- *MUST BE* '1' for Dn/An & (An)+
              END IF;

              IF (mode = mode_indir) OR (mode = mode_indir_pd) OR (mode = mode_indir_pi) OR (mode = mode_indir_dsp) OR
                 ((mode = mode_extend) AND ((data = mode_ext_addr) OR (data = mode_ext_imm))) THEN
                ram_en <= '1'; -- *MUST BE* '1' for (An), -(An), (An)+, d(An) & Addr
                ram_we <= '1'; -- *MUST BE* '1' for (An), -(An), (An)+, d(An) & Addr
              END IF;

              IF (mode = mode_extend) AND NOT((data = mode_ext_addr) OR (data = mode_ext_imm)) THEN
                port_we <= '1'; -- *MUST BE* '1' for Pn
              END IF;

            END IF;

            state <= st_save_writeback;

          WHEN st_save_writeback =>

            --
            -- Default Signal Assigns.
            --
            save_int_level <= '0';
            sel_ram_addr <= sraddr_pc;
            sel_register <= "1111";
            sel_reg_disp <= srdsp_one;
            sel_disp_sub <= sdsub_add;
            sel_feedback <= sfeed_reg;
            result_we <= '0';
            operand <= '0';

            --
            -- Save Interrupt Level in Control Register.
            --
            IF (save_int_level = '1') THEN
              int_we <= '1';
            END IF;

            --
            -- Create Quick (Extention) Data. 
            --
            IF (fetch = '1') THEN
              IF (((pre_inst = inst_bra) OR (pre_inst = inst_bsr)) AND
                 ((ram_dout(7) = '0') OR (ram_dout(7 DOWNTO 6) = "11"))) OR
                 (pre_inst = inst_load) THEN -- BraQ/BsrQ/LoadQ
                quick_data_ext <= ram_dout(7 DOWNTO 3);
              ELSE
                IF (ram_dout(11 DOWNTO 9) = mode_imm_quick) THEN -- ImmQ (Ext)
                  IF (pre_inst = inst_shl) OR (pre_inst = inst_shr) OR
                     (pre_inst = inst_add) OR (pre_inst = inst_sub) OR
                     (pre_inst = inst_mulu) OR (pre_inst = inst_divu) THEN -- ImmQ (Ext) Range is 8..15
                    quick_data_ext <= "00001";
                  ELSE -- ImmQ (-8..-1)
                    quick_data_ext <= "11111";
                  END IF;
                ELSE -- ImmQ (0..7)
                  quick_data_ext <= "00000";
                END IF;
              END IF;
            ELSE
              quick_data_ext <= "00000";
            END IF;

            --
            -- Interrupt Control Mechanism.
            --
            CASE int_stage IS

              WHEN int_none =>

                IF (fetch = '1') THEN
                  IF (ram_dout = inst_bra & branch_alw & "10" & mode_extend & "101") THEN -- RTI (Complete/Full Decode!)
                    int_stage <= int_pop_cr;
                  ELSIF (ram_dout(15) = '0') AND (ram_dout(11 DOWNTO 9) = "110") AND (ram_dout(5 DOWNTO 3) = "110") THEN -- PushL/PushU
                    int_stage <= int_push;
                    reg_mask <= ram_dout(13 DOWNTO 12) & ram_dout(8 DOWNTO 6) & ram_dout(2 DOWNTO 1);
                    reg_num <= "001";
                  ELSIF (ram_dout(15) = '0') AND (ram_dout(11 DOWNTO 6) = "111110") THEN -- PopL/PopU
                    int_stage <= int_pop;
                    reg_mask <= ram_dout(13 DOWNTO 12) & ram_dout(5 DOWNTO 1);
                    reg_num <= "001";
                  END IF;
                END IF;

              WHEN int_push_cr =>

                int_stage <= int_push_pc;

              WHEN int_push | int_pop =>

                reg_mask <= '0' & reg_mask(6 DOWNTO 1);
                reg_num <= reg_num + 1;

                IF (reg_num = "111") THEN
                  int_stage <= int_none;
                END IF;

              WHEN OTHERS =>

                int_stage <= int_none;

            END CASE;

            --
            -- Detect end of interrupt and trigger early detection of next.
            --
            IF (int_stage = int_pop_cr) THEN
              early_int_check <= '1';
            ELSE
              early_int_check <= '0';
            END IF;

            --
            -- Reshape Fetched Instruction & Create Interrupt/Push/Pop Instructions.
            --
            CASE int_stage IS

              WHEN int_none =>

                CASE pre_inst IS

                  WHEN inst_bra | inst_bsr =>

                    IF (ram_dout(7 DOWNTO 6) /= "10") THEN -- BraQ/BsrQ
                      instruction <= inst_add & mode_dir_areg & "111" &  mode_imm_quick & ram_dout(2 DOWNTO 0); -- AddQ #n,PC
                      inst_eq_bra_bsr <= '1';
                      inst_eq_move <= '0';
                    ELSIF (ram_dout(5 DOWNTO 0) = mode_extend & mode_ext_addr) THEN -- Bra/Bsr Relative
                      instruction <= inst_add & mode_dir_areg & "111" &  mode_extend & mode_ext_imm; -- Add #n,PC
                      inst_eq_bra_bsr <= '1';
                      inst_eq_move <= '0';
                    ELSIF (ram_dout(5 DOWNTO 0) = mode_extend & "101") THEN -- RTI (Incomplete/Fast Decode!)
                      instruction <= inst_move & mode_dir_areg & "111" & mode_indir_pi & "110"; -- Move (SP)+,PC
                      inst_eq_bra_bsr <= '0';
                      inst_eq_move <= '1';
                    ELSE
                      instruction <= inst_move & mode_dir_areg & "111" & ram_dout(5 DOWNTO 0); -- Move <src>,PC
                      inst_eq_bra_bsr <= '1';
                      inst_eq_move <= '1';
                    END IF;

                  WHEN inst_load =>

                    instruction <= ram_dout(15 DOWNTO 12) & "00" & ram_dout(11 DOWNTO 8) & mode_imm_quick & ram_dout(2 DOWNTO 0);
                    inst_eq_bra_bsr <= '0';
                    inst_eq_move <= '0';

                  WHEN OTHERS =>

                    IF (ram_dout(15) = '0') AND (ram_dout(11 DOWNTO 9) = "110") AND (ram_dout(5 DOWNTO 3) = "110") THEN -- PushL/PushU
                      instruction <= inst_move & mode_indir_pd & "110" & "000" & ram_dout(14) & "00"; -- Move Reg,-(SP)
                    ELSIF (ram_dout(15) = '0') AND (ram_dout(11 DOWNTO 6) = "111110") THEN -- PopL/PopU
                      instruction <= inst_move & "001" & ram_dout(14) & "11" & mode_indir_pi & "110"; -- Move (SP)+,Reg
                    ELSIF (ram_dout(11 DOWNTO 9) = mode_imm_quick) THEN
                      instruction <= ram_dout(15 DOWNTO 12) & ram_dout(5 DOWNTO 0) & mode_imm_quick & ram_dout(8 DOWNTO 6);
                    ELSE
                      instruction <= ram_dout;
                    END IF;
                    inst_eq_bra_bsr <= '0';
                    IF (pre_inst = inst_move) THEN
                      inst_eq_move <= '1';
                    ELSE
                      inst_eq_move <= '0';
                    END IF;

                END CASE;

              WHEN int_pop_cr =>

                instruction <= inst_move & mode_extend & "101" & mode_indir_pi & "110"; -- Move (SP)+,CR
                inst_eq_bra_bsr <= '0';
                inst_eq_move <= '1';

              WHEN int_push_cr =>

                instruction <= inst_move & mode_indir_pd & "110" & mode_extend & "101"; -- Move CR,-(SP)
                save_int_level <= '1';
                inst_eq_bra_bsr <= '0';
                inst_eq_move <= '1';

              WHEN int_push_pc =>

                instruction <= inst_move & mode_dir_areg & "111" & mode_imm_quick & int_level; -- Move PC,-(SP) & MoveQ #Int,PC
                inst_eq_bra_bsr <= '1';
                inst_eq_move <= '1';

              WHEN int_push =>

                instruction <= inst_move & mode_indir_pd & "110" & "00" &reg_num(2) & instruction(2) & reg_num(1 DOWNTO 0); -- Move Reg,-(SP)
                inst_eq_bra_bsr <= '0';
                inst_eq_move <= '1';

              WHEN int_pop =>

                instruction <= inst_move & "00" & NOT(reg_num(2)) & instruction(8) & NOT(reg_num(1 DOWNTO 0)) & mode_indir_pi & "110"; -- Move (SP)+,Reg
                inst_eq_bra_bsr <= '0';
                inst_eq_move <= '1';

            END CASE;

            --
            -- Detect Compare Instruction.
            --
            IF (fetch = '1') AND (pre_inst = inst_cmp) THEN
              inst_eq_cmp <= '1';
            ELSE
              inst_eq_cmp <= '0';
            END IF;

            --
            -- Detect Branch To Subroutine Instruction.
            --
            IF ((fetch = '1') AND (pre_inst = inst_bsr)) OR (int_stage = int_push_pc) THEN
              bsr <= '1';
            ELSE
              bsr <= '0';
            END IF;

            --
            -- Detect Invalid (Non-Executable) Instruction.
            --
            IF (fetch = '1') THEN
              CASE pre_inst IS

                WHEN inst_move | inst_and | inst_or | inst_xor | inst_shl |
                     inst_shr | inst_add | inst_sub =>
                  IF (ram_dout(11 DOWNTO 9) = mode_imm_quick) AND (ram_dout(5 DOWNTO 0) = mode_extend & mode_ext_imm) THEN
                    bad_inst <= '1';
                  ELSE
                    bad_inst <= '0';
                  END IF;

                WHEN inst_cmp =>

                  IF ((ram_dout(11 DOWNTO 9) = mode_imm_quick) AND (ram_dout(5 DOWNTO 3) = mode_imm_quick)) OR
                     ((ram_dout(11 DOWNTO 9) = mode_imm_quick) AND (ram_dout(5 DOWNTO 0) = mode_extend & mode_ext_imm)) OR
                     (ram_dout(11 DOWNTO 6) = mode_extend & mode_ext_imm) THEN
                    bad_inst <= '1';
                  ELSE
                    bad_inst <= '0';
                  END IF;

                WHEN inst_bra =>
                  IF (ram_dout(11 DOWNTO 8) /= "0000") AND (ram_dout(7 DOWNTO 0) = "10" & mode_extend & "101") THEN
                    bad_inst <= '1';
                  ELSE
                    bad_inst <= '0';
                  END IF;
                WHEN inst_bsr =>
                  IF (ram_dout(7 DOWNTO 0) = "10" & mode_extend & "101") THEN
                    bad_inst <= '1';
                  ELSE
                    bad_inst <= '0';
                  END IF;
                WHEN inst_load =>
                  bad_inst <= '0';
                WHEN OTHERS =>
                  bad_inst <= '1';
              END CASE;
            ELSE
              bad_inst <= '0';
            END IF;

            --
            -- Move to Fetch or Decode State.
            --
            IF (fetch = '1') OR (int_stage /= int_none) THEN -- Instruction Fetched/Created - Decode It
              fetch <= '0';
              state <= st_decode;
            ELSE -- Instruction Not Fetched - Fetch It
              fetch <= '1';
              ram_en <= '1';
              register_we <= '1';
              state <= st_fetch_execute;
            END IF;

        END CASE;

      END IF;
    END IF;
  END PROCESS;

  ---====================================---
  --== Interrupt Level Priority Encoder ==--
  ---====================================---

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN
        int_level <= "000";
      ELSIF (int_req7 = '1') AND (con_reg(6) = '1') THEN
        int_level <= "111";
      ELSIF (int_req6 = '1') AND (con_reg(5) = '1') THEN
        int_level <= "110";
      ELSIF (int_req5 = '1') AND (con_reg(4) = '1') THEN
        int_level <= "101";
      ELSIF (int_req4 = '1') AND (con_reg(3) = '1') THEN
        int_level <= "100";
      ELSIF (int_req3 = '1') AND (con_reg(2) = '1') THEN
        int_level <= "011";
      ELSIF (int_req2 = '1') AND (con_reg(1) = '1') THEN
        int_level <= "010";
      ELSIF (int_req1 = '1') AND (con_reg(0) = '1') THEN
        int_level <= "001";
      ELSE
        int_level <= "000";
      END IF;
    END IF;
  END PROCESS;

  ---=========================---
  --== Interrupt Acknowledge ==--
  ---=========================---

  Goi : FOR i IN 1 TO 7 GENERATE
    PROCESS(clk)
    BEGIN
      IF RISING_EDGE(clk) THEN
        IF (rst = '1') THEN
          int_ack_i(i) <= '0';
        ELSIF (state = st_save_writeback) AND (int_stage = int_pop_cr) AND
             (i = con_reg(14 DOWNTO 12)) THEN
          int_ack_i(i) <= '1';
        ELSE
          int_ack_i(i) <= '0';
        END IF;
      END IF;
    END PROCESS;
  END GENERATE Goi;

  int_ack1 <= int_ack_i(1);
  int_ack2 <= int_ack_i(2);
  int_ack3 <= int_ack_i(3);
  int_ack4 <= int_ack_i(4);
  int_ack5 <= int_ack_i(5);
  int_ack6 <= int_ack_i(6);
  int_ack7 <= int_ack_i(7);

  ---=======================---
  --== RAM Din Multiplexor ==--
  ---=======================---

  PROCESS(sel_ram_din, result, pc)
  BEGIN
    CASE sel_ram_din IS
      WHEN srdin_res => ram_din <= result;
      WHEN srdin_pc  => ram_din <= pc;
      WHEN OTHERS    => NULL;
    END CASE;
  END PROCESS;

  ---=============================---
  --== Data Register Multiplexor ==--
  ---=============================---

  PROCESS(sel_register, dregs)
  VARIABLE tmp : STD_LOGIC_VECTOR(2 DOWNTO 0);
  BEGIN
    tmp := sel_register(2 DOWNTO 0);
    CASE tmp IS
      WHEN "000" => cur_dreg <= dregs(0);
      WHEN "001" => cur_dreg <= dregs(1);
      WHEN "010" => cur_dreg <= dregs(2);
      WHEN "011" => cur_dreg <= dregs(3);
      WHEN "100" => cur_dreg <= dregs(4);
      WHEN "101" => cur_dreg <= dregs(5);
      WHEN "110" => cur_dreg <= dregs(6);
      WHEN "111" => cur_dreg <= dregs(7);
      WHEN OTHERS => NULL;
    END CASE;
  END PROCESS;

  ---================================---
  --== Address Register Multiplexor ==--
  ---================================---

  PROCESS(sel_register, aregs)
  VARIABLE tmp : STD_LOGIC_VECTOR(2 DOWNTO 0);
  BEGIN
    tmp := sel_register(2 DOWNTO 0);
    CASE tmp IS
      WHEN "000" => cur_areg <= aregs(0);
      WHEN "001" => cur_areg <= aregs(1);
      WHEN "010" => cur_areg <= aregs(2);
      WHEN "011" => cur_areg <= aregs(3);
      WHEN "100" => cur_areg <= aregs(4);
      WHEN "101" => cur_areg <= aregs(5);
      WHEN "110" => cur_areg <= aregs(6);
      WHEN "111" => cur_areg <= aregs(7);
      WHEN OTHERS => NULL;
    END CASE;
  END PROCESS;

  ---==========================================---
  --== Source/Destination Operand Multiplexor ==--
  ---==========================================---

  PROCESS(sel_op_source, ram_dout, quick_data_ext, instruction, cur_dreg, cur_areg, cur_port_reg)
  BEGIN
    CASE sel_op_source IS
      WHEN sopsrc_ram   => op_source <= ram_dout;
      WHEN sopsrc_port  => op_source <= cur_port_reg;
      WHEN sopsrc_dreg  => op_source <= cur_dreg;
      WHEN sopsrc_areg  => op_source <= cur_areg;
      WHEN OTHERS       => op_source(15 DOWNTO 8) <= (OTHERS => quick_data_ext(4));
                           op_source(7 DOWNTO 0) <= quick_data_ext & instruction(2 DOWNTO 0);
    END CASE;
  END PROCESS;

  ---=====================================---
  --== Register Displacement Multiplexor ==--
  ---=====================================---

  PROCESS(sel_reg_disp, ram_dout)
  BEGIN
    CASE sel_reg_disp IS
      WHEN srdsp_one => displaced_op <= "0000000000000001";
      WHEN srdsp_ram => displaced_op <= ram_dout;
      WHEN OTHERS    => NULL;
    END CASE;
  END PROCESS;

  ---=====================================---
  --== Register Displacement Calculation ==--
  ---=====================================---

  PROCESS(sel_disp_sub, displaced_op, cur_areg)
  BEGIN
    IF (sel_disp_sub = sdsub_add) THEN
      displaced <= cur_areg + displaced_op;
    ELSE
      displaced <= cur_areg - displaced_op;
    END IF;
  END PROCESS;

  ---=====================================---
  --== Address Hold Register Multiplexor ==--
  ---=====================================---

  PROCESS(sel_hold_reg, displaced, ram_dout)
  BEGIN
    CASE sel_hold_reg IS
      WHEN shreg_ram => hold_reg_d <= ram_dout;
      WHEN shreg_dsp => hold_reg_d <= displaced;
      WHEN OTHERS    => NULL;
    END CASE;
  END PROCESS;

  ---=========================---
  --== Address Hold Register ==--
  ---=========================---

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN
        hold_reg <= (OTHERS => '0'); 
      ELSIF (hold_we = '1') THEN
        hold_reg <= hold_reg_d;
      END IF;
    END IF;
  END PROCESS;

  ---===========================---
  --== RAM Address Multiplexor ==--
  ---===========================---

  PROCESS(sel_ram_addr, ram_dout, pc, sp, hold_reg, cur_areg)
  BEGIN
    CASE sel_ram_addr IS
      WHEN sraddr_pc   => ram_addr <= pc;
      WHEN sraddr_hold => ram_addr <= hold_reg;
      WHEN sraddr_sp   => ram_addr <= sp;
      WHEN sraddr_reg  => ram_addr <= cur_areg;
      WHEN OTHERS      => NULL;
    END CASE;
  END PROCESS;

  ---=======================================---
  --== Port Registers & Read/Write Signals ==--
  ---=======================================---

  Go : FOR i IN 0 TO 4 GENERATE
    PROCESS(clk)
    BEGIN
      IF RISING_EDGE(clk) THEN
        IF (rst = '1') THEN
          port_registers(i) <= (OTHERS => '0');
          port_write(i) <= '0';
        ELSIF (port_we = '1') AND (sel_port = i) THEN
          port_registers(i) <= result;
          port_write(i) <= '1';
        ELSE
          port_registers(i) <= port_registers(i);
          port_write(i) <= '0';
        END IF;
      END IF;
    END PROCESS;

    PROCESS(clk)
    BEGIN
      IF RISING_EDGE(clk) THEN
        IF (rst = '1') THEN
          port_read(i) <= '0';
        ELSIF (state = st_decode) AND (operand = '0') AND (bad_inst = '0') AND
              (mode = mode_extend) AND (data = i) THEN
          port_read(i) <= '1';
        ELSE
          port_read(i) <= '0';
        END IF;
      END IF;
    END PROCESS;

  END GENERATE Go;

  port0_dout <= port_registers(0);
  port1_dout <= port_registers(1);
  port2_dout <= port_registers(2);
  port3_dout <= port_registers(3);
  port4_dout <= port_registers(4);

  port0_write <= port_write(0);
  port1_write <= port_write(1);
  port2_write <= port_write(2);
  port3_write <= port_write(3);
  port4_write <= port_write(4);

  port0_read <= port_read(0);
  port1_read <= port_read(1);
  port2_read <= port_read(2);
  port3_read <= port_read(3);
  port4_read <= port_read(4);

  ---=====================================---
  --== Port/Control Register Multiplexor ==--
  ---=====================================---

  PROCESS(sel_port, port0_din, port1_din, port2_din, port3_din, port4_din, con_reg)
  BEGIN
    CASE sel_port IS
      WHEN "000" => cur_port_reg <= port0_din;
      WHEN "001" => cur_port_reg <= port1_din;
      WHEN "010" => cur_port_reg <= port2_din;
      WHEN "011" => cur_port_reg <= port3_din;
      WHEN "100" => cur_port_reg <= port4_din;
      WHEN "101" => cur_port_reg <= con_reg;
      WHEN "110" => cur_port_reg <= (OTHERS => '0'); -- TBD
      WHEN "111" => cur_port_reg <= (OTHERS => '0'); -- TBD
      WHEN OTHERS => NULL;
    END CASE;
  END PROCESS;

  ---==================---
  --== Data Registers ==--
  ---==================---

  GDRegs : FOR i IN 0 TO 7 GENERATE
    PROCESS(clk)
    BEGIN
      IF RISING_EDGE(clk) THEN
        IF (rst = '1') THEN
          dregs(i) <= (OTHERS => '0');
        ELSIF (register_we = '1') AND (sel_register(3) = '0') AND
              (sel_register(2 DOWNTO 0) = i) THEN
          dregs(i) <= result;
        END IF;
      END IF;
    END PROCESS;
  END GENERATE GDRegs;

  ---=====================---
  --== Address Registers ==--
  ---=====================---

  GARegs : FOR i IN 0 TO 7 GENERATE
    PROCESS(clk)
    BEGIN
      IF RISING_EDGE(clk) THEN
        IF (rst = '1') THEN
          aregs(i) <= (OTHERS => '0');
        ELSIF (register_we = '1') AND (sel_register(3) = '1') AND
              (sel_register(2 DOWNTO 0) = i) THEN
          aregs(i) <= feedback;
        END IF;
      END IF;
    END PROCESS;
  END GENERATE GARegs;

  ---=================================---
  --== Register Feedback Multiplexor ==--
  ---=================================---

  PROCESS(sel_feedback, result, displaced)
  BEGIN
    CASE sel_feedback IS
      WHEN sfeed_reg => feedback <= displaced;
      WHEN sfeed_res => feedback <= result;
      WHEN OTHERS    => NULL;
    END CASE;
  END PROCESS;

  ---===========================---
  --== Source Operand Register ==--
  ---===========================---

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN
        src_op <= (OTHERS => '0'); 
      ELSIF (src_op_we = '1') THEN
        src_op <= op_source;
      END IF;
    END IF;
  END PROCESS;

  ---================================---
  --== Destination Operand Register ==--
  ---================================---

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN
        dst_op <= (OTHERS => '0'); 
      ELSIF (dst_op_we = '1') THEN
        dst_op <= op_source;
      END IF;
    END IF;
  END PROCESS;

  ---==================================---
  --== Left Shift Combinational Logic ==--
  ---==================================---

  PROCESS(src_op, dst_op)
  VARIABLE tmp : STD_LOGIC_VECTOR(3 DOWNTO 0);
  BEGIN
    tmp := src_op(3 DOWNTO 0);
    CASE tmp IS
      WHEN "0000" => result_shl <= dst_op(15 DOWNTO 00);
      WHEN "0001" => result_shl <= dst_op(14 DOWNTO 00) & '0';
      WHEN "0010" => result_shl <= dst_op(13 DOWNTO 00) & "00";
      WHEN "0011" => result_shl <= dst_op(12 DOWNTO 00) & "000";
      WHEN "0100" => result_shl <= dst_op(11 DOWNTO 00) & "0000";
      WHEN "0101" => result_shl <= dst_op(10 DOWNTO 00) & "00000";
      WHEN "0110" => result_shl <= dst_op(09 DOWNTO 00) & "000000";
      WHEN "0111" => result_shl <= dst_op(08 DOWNTO 00) & "0000000";
      WHEN "1000" => result_shl <= dst_op(07 DOWNTO 00) & "00000000";
      WHEN "1001" => result_shl <= dst_op(06 DOWNTO 00) & "000000000";
      WHEN "1010" => result_shl <= dst_op(05 DOWNTO 00) & "0000000000";
      WHEN "1011" => result_shl <= dst_op(04 DOWNTO 00) & "00000000000";
      WHEN "1100" => result_shl <= dst_op(03 DOWNTO 00) & "000000000000";
      WHEN "1101" => result_shl <= dst_op(02 DOWNTO 00) & "0000000000000";
      WHEN "1110" => result_shl <= dst_op(01 DOWNTO 00) & "00000000000000";
      WHEN "1111" => result_shl <= dst_op(00 DOWNTO 00) & "000000000000000";
      WHEN OTHERS => NULL;
    END CASE;
  END PROCESS;

  ---===================================---
  --== Right Shift Combinational Logic ==--
  ---===================================---

  PROCESS(src_op, dst_op)
  VARIABLE tmp : STD_LOGIC_VECTOR(3 DOWNTO 0);
  BEGIN
    tmp := src_op(3 DOWNTO 0);
    CASE tmp IS
      WHEN "0000" => result_shr <= dst_op(15 DOWNTO 00);
      WHEN "0001" => result_shr <= '0' & dst_op(15 DOWNTO 01);
      WHEN "0010" => result_shr <= "00" & dst_op(15 DOWNTO 02);
      WHEN "0011" => result_shr <= "000" & dst_op(15 DOWNTO 03);
      WHEN "0100" => result_shr <= "0000" & dst_op(15 DOWNTO 04);
      WHEN "0101" => result_shr <= "00000" & dst_op(15 DOWNTO 05);
      WHEN "0110" => result_shr <= "000000" & dst_op(15 DOWNTO 06);
      WHEN "0111" => result_shr <= "0000000" & dst_op(15 DOWNTO 07);
      WHEN "1000" => result_shr <= "00000000" & dst_op(15 DOWNTO 08);
      WHEN "1001" => result_shr <= "000000000" & dst_op(15 DOWNTO 09);
      WHEN "1010" => result_shr <= "0000000000" & dst_op(15 DOWNTO 10);
      WHEN "1011" => result_shr <= "00000000000" & dst_op(15 DOWNTO 11);
      WHEN "1100" => result_shr <= "000000000000" & dst_op(15 DOWNTO 12);
      WHEN "1101" => result_shr <= "0000000000000" & dst_op(15 DOWNTO 13);
      WHEN "1110" => result_shr <= "00000000000000" & dst_op(15 DOWNTO 14);
      WHEN "1111" => result_shr <= "000000000000000" & dst_op(15 DOWNTO 15);
      WHEN OTHERS => NULL;
    END CASE;
  END PROCESS;

  ---===================================---
  --== Condition Code Generation Logic ==--
  ---===================================---

  PROCESS(inst, src_op, dst_op, result)
  BEGIN
    IF (result = "0000000000000000") THEN
      zr_i <= '1';
    ELSE
      zr_i <= '0';
    END IF;
    ng_i <= result(15);
    CASE inst IS
      WHEN inst_add =>
        so_i <= (NOT(src_op(15)) AND NOT(dst_op(15)) AND result(15)) OR (src_op(15) AND dst_op(15) AND NOT(result(15)));
        uo_i <= (src_op(15) AND dst_op(15)) OR (src_op(15) AND NOT(result(15))) OR (dst_op(15) AND NOT(result(15)));
      WHEN inst_sub | inst_cmp =>
        so_i <= (NOT(src_op(15)) AND dst_op(15) AND NOT(result(15))) OR (src_op(15) AND NOT(dst_op(15)) AND result(15));
        uo_i <= (src_op(15) AND NOT(dst_op(15))) OR (src_op(15) AND result(15)) OR (NOT(dst_op(15)) AND result(15));
      WHEN OTHERS =>
        so_i <= '0';
        uo_i <= '0';
    END CASE;
  END PROCESS;

  ---==============================---
  --== Condition Code Multiplexor ==--
  ---==============================---

  PROCESS(block_conditions, zr, ng, so, uo, zr_i, ng_i, so_i, uo_i, port_we, sel_port, result)
  BEGIN
    IF (block_conditions = '1') THEN -- Used last registered conditions
      zr_m <= zr;
      ng_m <= ng;
      so_m <= so;
      uo_m <= uo;
    ELSIF (port_we = '1') AND (sel_port = 5) THEN -- Use new CR conditions
      zr_m <= result(8);
      ng_m <= result(9);
      so_m <= result(10);
      uo_m <= result(11);
    ELSE -- Use new conditions that are about to be registered
      zr_m <= zr_i;
      ng_m <= ng_i;
      so_m <= so_i;
      uo_m <= uo_i;
    END IF;
  END PROCESS;

  ---====================---
  --== Control Register ==--
  ---====================---

    PROCESS(clk)
    BEGIN
      IF RISING_EDGE(clk) THEN
        IF (rst = '1') THEN
          con_reg <= (OTHERS => '0');
        ELSIF (port_we = '1') AND (sel_port = 5) THEN
          con_reg <= result;
        ELSIF (int_we = '1') THEN
          con_reg(14 DOWNTO 12) <= int_level;
        ELSIF (result_we_delay = '1') AND (block_conditions = '0') THEN
          con_reg(11 DOWNTO 8) <= so_i & uo_i & ng_i & zr_i;
        END IF;
      END IF;
    END PROCESS;

  ---==================================================---
  --== Condition Code Check & Execution Determination ==--
  ---==================================================---

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN
        execute <= '0';
        result_we_delay <= '0';

      ELSE

        result_we_delay <= result_we;

        IF (state = st_save_writeback) THEN

          CASE int_stage IS
            WHEN int_none =>

              IF (fetch = '1') THEN
                CASE pre_inst IS
                  WHEN inst_bra | inst_bsr =>
                    CASE condition IS
                      WHEN branch_alw => execute <= '1';
                      WHEN branch_nev => execute <= '0';
                      WHEN branch_zrs => execute <= zr_m;
                      WHEN branch_zrc => execute <= NOT(zr_m);
                      WHEN branch_ngs => execute <= ng_m;
                      WHEN branch_ngc => execute <= NOT(ng_m);
                      WHEN branch_uos => execute <= uo_m;
                      WHEN branch_uoc => execute <= NOT(uo_m);
                      WHEN branch_sos => execute <= so_m;
                      WHEN branch_soc => execute <= NOT(so_m);
                      WHEN branch_ugt => execute <= NOT(uo_m OR zr_m);
                      WHEN branch_ule => execute <= uo_m OR zr_m;
                      WHEN branch_sgt => execute <= NOT((so_m XOR ng_m) OR zr_m);
                      WHEN branch_sge => execute <= NOT(so_m XOR ng_m);
                      WHEN branch_slt => execute <= so_m XOR ng_m;
                      WHEN branch_sle => execute <= (so_m XOR ng_m) OR zr_m;
                      WHEN OTHERS => NULL;
                    END CASE;
                  WHEN inst_move | inst_and | inst_or | inst_xor | inst_shl | inst_shr | inst_add | inst_sub =>
                    IF ((ram_dout(11 DOWNTO 9) = "110") AND (ram_dout(5 DOWNTO 3) = "110")) OR -- PushL/PushU
                       (ram_dout(11 DOWNTO 6) = "111110") THEN -- PopL/PopU
                      execute <= ram_dout(0);
                    ELSE
                      execute <= '1';
                    END IF;
                  WHEN OTHERS =>
                    execute <= '1';
                END CASE;
              ELSE
                execute <= '0'; -- Used when fetch is delayed (i.e. dst=PC)
              END IF;
            WHEN int_push | int_pop =>
              execute <= reg_mask(0);
            WHEN OTHERS =>
              execute <= '1';
          END CASE;

        END IF;
      END IF;
    END IF;
  END PROCESS;

  ---=============================---
  --== Arithmetic and Logic Unit ==--
  ---=============================---

  PROCESS(src_op, dst_op, result_shl, result_shr, inst)
  BEGIN
    CASE inst IS
      WHEN inst_move | inst_load => result_i <= src_op;
      WHEN inst_and  => result_i <= src_op AND dst_op;
      WHEN inst_or   => result_i <= src_op OR dst_op;
      WHEN inst_xor  => result_i <= src_op XOR dst_op;
      WHEN inst_shl  => result_i <= result_shl;
      WHEN inst_shr  => result_i <= result_shr;
      WHEN inst_add  => result_i <= dst_op + src_op;
      WHEN OTHERS    => result_i <= dst_op - src_op;
    END CASE;
  END PROCESS;

  ---===============================---
  --== Instruction Result Register ==--
  ---===============================---

  PROCESS(clk)
  BEGIN
    IF RISING_EDGE(clk) THEN
      IF (rst = '1') THEN
        result <= (OTHERS => '0');
      ELSIF (result_we = '1') THEN
        result <= result_i;
      END IF;
    END IF;
  END PROCESS;

END rtl;
